@@ -23,7 +23,7 @@
  *  doi:10.1109/TKDE.2016.2551240
  *
  *
- *  Copyleft (C) 2018-2025, Marek Gagolewski <https://www.gagolewski.com>
+ *  Copyleft (C) 2018-2026, Marek Gagolewski <https://www.gagolewski.com>
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU Affero General Public License
@@ -409,9 +409,6 @@ CComparePartitionsInfoResult Ccompare_partitions_info(const T* C,
  *  with columns permuted based on the solution to the
  *  maximal linear sum assignment problem.
  *
- *  For non-square matrices, missing rows/columns are assumed
- *  to be filled with 0s.
- *
  *  Accuracy(C[:,sigma]) is sometimes referred to as
  *  set-matching classification rate or pivoted accuracy.
  *
@@ -440,6 +437,8 @@ CComparePartitionsInfoResult Ccompare_partitions_info(const T* C,
 template<class T>
 double Ccompare_partitions_npa(const T* C, Py_ssize_t xc, Py_ssize_t yc)
 {
+    GENIECLUST_ASSERT(xc == yc);  // TODO
+
     double n = 0.0; // total sum (length of the underlying x and y = number of points)
     for (Py_ssize_t ij=0; ij<xc*yc; ++ij) {
         if (C[ij] > 0) {
@@ -449,6 +448,7 @@ double Ccompare_partitions_npa(const T* C, Py_ssize_t xc, Py_ssize_t yc)
 
     // if C is not a square matrix, treat the missing columns
     // as if they were filled with 0s
+    // TODO This is not a good strategy!
     Py_ssize_t xyc = std::max(xc, yc);
     std::vector<double> S(xyc*xyc, 0.0);
     for (Py_ssize_t i=0; i<xc; ++i) {
@@ -482,10 +482,6 @@ double Ccompare_partitions_npa(const T* C, Py_ssize_t xc, Py_ssize_t yc)
  *  NCA is not symmetric - we assume that rows in C determine the true
  *  (reference) partition.
  *
- *  For non-square confusion matrices, missing rows/columns
- *  are assumed to be filled with 0s and that 0/0 is 0,
- *  but the original row count is used for normalisation.
- *
  *  References
  *  ==========
  *
@@ -503,6 +499,8 @@ double Ccompare_partitions_npa(const T* C, Py_ssize_t xc, Py_ssize_t yc)
 template<class T>
 double Ccompare_partitions_nca(const T* C, Py_ssize_t xc, Py_ssize_t yc)
 {
+    GENIECLUST_ASSERT(xc == yc);  // TODO
+
     std::vector<double> sum_x(xc, 0.0);
     for (Py_ssize_t i=0; i<xc; ++i) {
         for (Py_ssize_t j=0; j<yc; ++j) {
@@ -512,12 +510,15 @@ double Ccompare_partitions_nca(const T* C, Py_ssize_t xc, Py_ssize_t yc)
         }
     }
 
+
+
     // if xc>yc, treat C as if its missing columns were filled with 0s
     Py_ssize_t yc2 = std::max(xc, yc);
 
     // if xc<yc, only xc items are matched;
     // thus, overall, the behaviour is like filling missed rows/columns with 0s
     // and assuming 0/0 == 0, while still using k=nrow(C)
+    // TODO This is not a good strategy!
 
     std::vector<double> S(xc*yc2, 0.0);
     for (Py_ssize_t i=0; i<xc; ++i) {
